@@ -19,6 +19,9 @@
 #include "sbd.h"
 #include <sys/reboot.h>
 #include <sys/types.h>
+#ifdef __GLIBC__
+#include <sys/sysmacros.h>
+#endif
 #include <sys/stat.h>
 #include <pwd.h>
 #include <unistd.h>
@@ -44,7 +47,7 @@ int		timeout_startup		= 120;
 
 int	watchdog_use		= 1;
 int	watchdog_set_timeout	= 1;
-unsigned long	timeout_watchdog_crashdump = 240;
+unsigned long	timeout_watchdog_crashdump = 0;
 int	skip_rt			= 0;
 int	debug			= 0;
 int	debug_mode		= 0;
@@ -78,7 +81,7 @@ usage(void)
 "-T		Do NOT initialize the watchdog timeout (watch only)\n"
 "-S <0|1>	Set start mode if the node was previously fenced (watch only)\n"
 "-p <path>	Write pidfile to the specified path (watch only)\n"
-"-v		Enable some verbose debug logging (optional)\n"
+"-v|-vv|-vvv	Enable verbose|debug|debug-library logging (optional)\n"
 "\n"
 "-1 <N>		Set watchdog timeout to N seconds (optional, create only)\n"
 "-2 <N>		Set slot allocation timeout to N seconds (optional, create only)\n"
@@ -86,7 +89,8 @@ usage(void)
 "-4 <N>		Set msgwait timeout to N seconds (optional, create only)\n"
 "-5 <N>		Warn if loop latency exceeds threshold (optional, watch only)\n"
 "			(default is 3, set to 0 to disable)\n"
-"-C <N>		Watchdog timeout to set before crashdumping (def: 240s, optional)\n"
+"-C <N>		Watchdog timeout to set before crashdumping\n"
+"			(def: 0s = disable gracefully, optional)\n"
 "-I <N>		Async IO read timeout (defaults to 3 * loop timeout, optional)\n"
 "-s <N>		Timeout to wait for devices to become available (def: 120s)\n"
 "-t <N>		Dampening delay before faulty servants are restarted (optional)\n"
@@ -95,6 +99,8 @@ usage(void)
 "			(default is 1, set to 0 to disable)\n"
 "-P		Check Pacemaker quorum and node health (optional, watch only)\n"
 "-Z		Enable trace mode. WARNING: UNSAFE FOR PRODUCTION!\n"
+"-r		Set timeout-action to comma-separated combination of\n"
+"		noflush|flush plus reboot|crashdump|off (default is flush,reboot)\n"
 "Commands:\n"
 #if SUPPORT_SHARED_DISK
 "create		initialize N slots on <dev> - OVERWRITES DEVICE!\n"
@@ -102,7 +108,7 @@ usage(void)
 "dump		Dump meta-data header from device.\n"
 "allocate <node>\n"
 "		Allocate a slot for node (optional)\n"
-"message <node> (test|reset|off|clear|exit)\n"
+"message <node> (test|reset|off|crashdump|clear|exit)\n"
 "		Writes the specified message to node's slot.\n"
 #endif
 "watch		Loop forever, monitoring own slot\n"
@@ -248,7 +254,8 @@ watchdog_close(bool disarm)
 #define MAX_WATCHDOGS 64
 #define SYS_CLASS_WATCHDOG "/sys/class/watchdog"
 #define SYS_CHAR_DEV_DIR "/sys/dev/char"
-#define WATCHDOG_NODEDIR "/dev"
+#define WATCHDOG_NODEDIR "/dev/"
+#define WATCHDOG_NODEDIR_LEN 5
 
 struct watchdog_list_item {
 	dev_t dev;
@@ -258,6 +265,12 @@ struct watchdog_list_item {
 	struct watchdog_list_item *next;
 };
 
+struct link_list_item {
+	char *dev_node;
+	char *link_name;
+	struct link_list_item *next;
+};
+
 static struct watchdog_list_item *watchdog_list = NULL;
 static int watchdog_list_items = 0;
 
@@ -268,9 +281,10 @@ watchdog_populate_list(void)
 		{makedev(10,130), 0};
 	int num_watchdogs = 1;
 	struct dirent *entry;
-	char entry_name[64];
+	char entry_name[280];
 	DIR *dp;
-	char buf[256] = "";
+	char buf[280] = "";
+	struct link_list_item *link_list = NULL;
 
 	if (watchdog_list != NULL) {
 		return;
@@ -284,7 +298,7 @@ watchdog_populate_list(void)
 				FILE *file;
 
 				snprintf(entry_name, sizeof(entry_name),
-						SYS_CLASS_WATCHDOG "/%s/dev", entry->d_name);
+				         SYS_CLASS_WATCHDOG "/%s/dev", entry->d_name);
 				file = fopen(entry_name, "r");
 				if (file) {
 					int major, minor;
@@ -305,12 +319,59 @@ watchdog_populate_list(void)
 	/* search for watchdog nodes in /dev */
 	dp = opendir(WATCHDOG_NODEDIR);
 	if (dp) {
+		/* first go for links and memorize them */
 		while ((entry = readdir(dp))) {
-			if ((entry->d_type == DT_CHR) || (entry->d_type == DT_LNK)) {
+			if (entry->d_type == DT_LNK) {
+				int len;
+
+				snprintf(entry_name, sizeof(entry_name),
+				         WATCHDOG_NODEDIR "%s", entry->d_name);
+
+				/* !realpath(entry_name, buf) unfortunately does a stat on
+				 * target so we can't really use it to check if links stay
+				 * within /dev without triggering e.g. AVC-logs (with
+				 * SELinux policy that just allows stat within /dev).
+				 * Without canonicalization that doesn't actually touch the
+				 * filesystem easily available introduce some limitations
+				 * for simplicity:
+				 * - just simple path without '..'
+				 * - just one level of symlinks (avoid e.g. loop-checking)
+				 */
+				len = readlink(entry_name, buf, sizeof(buf) - 1);
+				if ((len < 1) ||
+				    (len > sizeof(buf) - WATCHDOG_NODEDIR_LEN - 1)) {
+					continue;
+				}
+				buf[len] = '\0';
+				if (buf[0] != '/') {
+					memmove(&buf[WATCHDOG_NODEDIR_LEN], buf, len+1);
+					memcpy(buf, WATCHDOG_NODEDIR, WATCHDOG_NODEDIR_LEN);
+					len += WATCHDOG_NODEDIR_LEN;
+				}
+				if (strstr(buf, "/../") ||
+				    strncmp(WATCHDOG_NODEDIR, buf, WATCHDOG_NODEDIR_LEN)) {
+					continue;
+				} else {
+					/* just memorize to avoid statting the target - SELinux */
+					struct link_list_item *lli =
+						calloc(1, sizeof(struct link_list_item));
+
+					lli->dev_node = strdup(buf);
+					lli->link_name = strdup(entry_name);
+					lli->next = link_list;
+					link_list = lli;
+				}
+			}
+		}
+
+		rewinddir(dp);
+
+		while ((entry = readdir(dp))) {
+			if (entry->d_type == DT_CHR) {
 				struct stat statbuf;
 
 				snprintf(entry_name, sizeof(entry_name),
-						WATCHDOG_NODEDIR "/%s", entry->d_name);
+				         WATCHDOG_NODEDIR "%s", entry->d_name);
 				if(!stat(entry_name, &statbuf) && S_ISCHR(statbuf.st_mode)) {
 					int i;
 
@@ -318,7 +379,9 @@ watchdog_populate_list(void)
 						if (statbuf.st_rdev == watchdogs[i]) {
 							int wdfd = watchdog_init_fd(entry_name, -1);
 							struct watchdog_list_item *wdg =
-									calloc(1, sizeof(struct watchdog_list_item));
+								calloc(1, sizeof(struct watchdog_list_item));
+							int len;
+							struct link_list_item *tmp_list = NULL;
 
 							wdg->dev = watchdogs[i];
 							wdg->dev_node = strdup(entry_name);
@@ -338,23 +401,60 @@ watchdog_populate_list(void)
 							}
 
 							snprintf(entry_name, sizeof(entry_name),
-								SYS_CHAR_DEV_DIR "/%d:%d/device/driver",
-								major(watchdogs[i]), minor(watchdogs[i]));
-							if (readlink(entry_name, buf, sizeof(buf)) > 0) {
+							         SYS_CHAR_DEV_DIR "/%d:%d/device/driver",
+							         major(watchdogs[i]), minor(watchdogs[i]));
+							len = readlink(entry_name, buf, sizeof(buf) - 1);
+							if (len > 0) {
+								buf[len] = '\0';
 								wdg->dev_driver = strdup(basename(buf));
 							} else if ((wdg->dev_ident) &&
-										(strcmp(wdg->dev_ident,
-												"Software Watchdog") == 0)) {
+							           (strcmp(wdg->dev_ident,
+							                   "Software Watchdog") == 0)) {
 								wdg->dev_driver = strdup("softdog");
 							}
+
+							/* create dupes if we have memorized links
+							 * to this node
+							 */
+							for (tmp_list = link_list; tmp_list;
+							     tmp_list = tmp_list->next) {
+								if (!strcmp(tmp_list->dev_node,
+								            wdg->dev_node)) {
+									struct watchdog_list_item *dupe_wdg =
+										calloc(1, sizeof(struct watchdog_list_item));
+
+									/* as long as we never purge watchdog_list
+									 * there is no need to dupe strings
+									 */
+									*dupe_wdg = *wdg;
+									dupe_wdg->dev_node = strdup(tmp_list->link_name);
+									dupe_wdg->next = watchdog_list;
+									watchdog_list = dupe_wdg;
+									watchdog_list_items++;
+								}
+								/* for performance reasons we could remove
+								 * the link_list entry
+								 */
+							}
 							break;
 						}
 					}
 				}
 			}
 		}
+
 		closedir(dp);
 	}
+
+	/* cleanup link list */
+	while (link_list) {
+		struct link_list_item *tmp_list = link_list;
+
+		link_list = link_list->next;
+		free(tmp_list->dev_node);
+		free(tmp_list->link_name);
+		free(tmp_list);
+	}
 }
 
 int watchdog_info(void)
@@ -469,13 +569,13 @@ enum {
 #define IOPRIO_PRIO_DATA(mask)  ((mask) & IOPRIO_PRIO_MASK)
 #define IOPRIO_PRIO_VALUE(class, data)  (((class) << IOPRIO_CLASS_SHIFT) | data)
 
-static unsigned char
+static void
 sbd_stack_hogger(unsigned char * inbuf, int kbytes)
 {
     unsigned char buf[1024];
 
     if(kbytes <= 0) {
-        return HOG_CHAR;
+        return;
     }
 
     if (inbuf == NULL) {
@@ -485,10 +585,10 @@ sbd_stack_hogger(unsigned char * inbuf, int kbytes)
     }
 
     if (kbytes > 0) {
-        return sbd_stack_hogger(buf, kbytes-1);
-    } else {
-        return buf[sizeof(buf)-1];
+        sbd_stack_hogger(buf, kbytes-1);
     }
+
+    return;
 }
 
 static void
@@ -563,6 +663,112 @@ static void sbd_memlock(int stackgrowK, int heapgrowK)
 #endif
 }
 
+static int get_realtime_budget(void)
+{
+    FILE *f;
+    char fname[PATH_MAX];
+    int res = -1, lnum = 0;
+    char *cgroup = NULL, *namespecs = NULL;
+
+    snprintf(fname, PATH_MAX, "/proc/%jd/cgroup", (intmax_t)getpid());
+    f = fopen(fname, "rt");
+    if (f == NULL) {
+        cl_log(LOG_WARNING, "Can't open cgroup file for pid=%jd",
+                            (intmax_t)getpid());
+        goto exit_res;
+    }
+    while( fscanf(f, "%d:%m[^:]:%m[^\n]", &lnum,  &namespecs, &cgroup) !=EOF ) {
+        if (namespecs && strstr(namespecs, "cpuacct")) {
+            free(namespecs);
+            break;
+        }
+        if (cgroup) {
+            free(cgroup);
+            cgroup = NULL;
+        }
+        if (namespecs) {
+            free(namespecs);
+            namespecs = NULL;
+        }
+    }
+    fclose(f);
+    if (cgroup == NULL) {
+        cl_log(LOG_WARNING, "Failed getting cgroup for pid=%jd",
+                            (intmax_t)getpid());
+        goto exit_res;
+    }
+    snprintf(fname, PATH_MAX, "/sys/fs/cgroup/cpu%s/cpu.rt_runtime_us",
+                              cgroup);
+    f = fopen(fname, "rt");
+    if (f == NULL) {
+        cl_log(LOG_WARNING, "cpu.rt_runtime_us existed for root-slice but "
+            "doesn't for '%s'", cgroup);
+        goto exit_res;
+    }
+    if (fscanf(f, "%d", &res) != 1) {
+        cl_log(LOG_WARNING, "failed reading rt-budget from %s", fname);
+    } else {
+        cl_log(LOG_INFO, "slice='%s' has rt-budget=%d", cgroup, res);
+    }
+    fclose(f);
+
+exit_res:
+    if (cgroup) {
+        free(cgroup);
+    }
+    return res;
+}
+
+/* stolen from corosync */
+static int sbd_move_to_root_cgroup(bool enforce_root_cgroup) {
+    FILE *f;
+    int res = -1;
+
+    /*
+     * /sys/fs/cgroup is hardcoded, because most of Linux distributions are now
+     * using systemd and systemd uses hardcoded path of cgroup mount point.
+     *
+     * This feature is expected to be removed as soon as systemd gets support
+     * for managing RT configuration.
+     */
+    f = fopen("/sys/fs/cgroup/cpu/cpu.rt_runtime_us", "rt");
+    if (f == NULL) {
+        cl_log(LOG_DEBUG, "cpu.rt_runtime_us doesn't exist -> "
+            "system without cgroup or with disabled CONFIG_RT_GROUP_SCHED");
+        res = 0;
+        goto exit_res;
+    }
+    fclose(f);
+
+    if ((!enforce_root_cgroup) && (get_realtime_budget() > 0)) {
+        cl_log(LOG_DEBUG, "looks as if we have rt-budget in the slice we are "
+                          "-> skip moving to root-slice");
+        res = 0;
+        goto exit_res;
+    }
+
+    f = fopen("/sys/fs/cgroup/cpu/tasks", "w");
+    if (f == NULL) {
+        cl_log(LOG_WARNING, "Can't open cgroups tasks file for writing");
+
+        goto exit_res;
+    }
+
+    if (fprintf(f, "%jd\n", (intmax_t)getpid()) <= 0) {
+        cl_log(LOG_WARNING, "Can't write sbd pid into cgroups tasks file");
+        goto close_and_exit_res;
+    }
+
+close_and_exit_res:
+    if (fclose(f) != 0) {
+        cl_log(LOG_WARNING, "Can't close cgroups tasks file");
+        goto exit_res;
+    }
+
+exit_res:
+    return (res);
+}
+
 void
 sbd_make_realtime(int priority, int stackgrowK, int heapgrowK)
 {
@@ -571,6 +777,10 @@ sbd_make_realtime(int priority, int stackgrowK, int heapgrowK)
     }
 
 #ifdef SCHED_RR
+    if (move_to_root_cgroup) {
+        sbd_move_to_root_cgroup(enforce_moving_to_root_cgroup);
+    }
+
     {
         int pcurrent = 0;
         int pmin = sched_get_priority_min(SCHED_RR);
@@ -672,7 +882,7 @@ sysrq_trigger(char t)
 
 
 static void
-do_exit(char kind) 
+do_exit(char kind, bool do_flush)
 {
     /* TODO: Turn debug_mode into a bit field? Delay + kdump for example */
     const char *reason = NULL;
@@ -681,12 +891,12 @@ do_exit(char kind)
         cl_log(LOG_NOTICE, "Initiating kdump");
 
     } else if (debug_mode == 1) {
-        cl_log(LOG_WARNING, "Initiating kdump instead of panicing the node (debug mode)");
+        cl_log(LOG_WARNING, "Initiating kdump instead of panicking the node (debug mode)");
         kind = 'c';
     }
 
     if (debug_mode == 2) {
-        cl_log(LOG_WARNING, "Shutting down SBD instead of panicing the node (debug mode)");
+        cl_log(LOG_WARNING, "Shutting down SBD instead of panicking the node (debug mode)");
         watchdog_close(true);
         exit(0);
     }
@@ -717,12 +927,21 @@ do_exit(char kind)
     }
 
     cl_log(LOG_EMERG, "Rebooting system: %s", reason);
-    sync();
+    if (do_flush) {
+        sync();
+    }
 
-    if(kind == 'c') {
-        watchdog_close(true);
+    if (kind == 'c') {
+        if (timeout_watchdog_crashdump) {
+            if (timeout_watchdog != timeout_watchdog_crashdump) {
+                timeout_watchdog = timeout_watchdog_crashdump;
+                watchdog_init_interval();
+            }
+            watchdog_close(false);
+        } else {
+            watchdog_close(true);
+        }
         sysrq_trigger(kind);
-
     } else {
         watchdog_close(false);
         sysrq_trigger(kind);
@@ -737,19 +956,25 @@ do_exit(char kind)
 void
 do_crashdump(void)
 {
-    do_exit('c');
+    do_exit('c', true);
 }
 
 void
 do_reset(void)
 {
-    do_exit('b');
+    do_exit('b', true);
 }
 
 void
 do_off(void)
 {
-    do_exit('o');
+    do_exit('o', true);
+}
+
+void
+do_timeout_action(void)
+{
+	do_exit(timeout_sysrq_char, do_flush);
 }
 
 /*
@@ -883,14 +1108,14 @@ notify_parent(void)
         /* Our parent died unexpectedly. Triggering
          * self-fence. */
         cl_log(LOG_WARNING, "Our parent is dead.");
-        do_reset();
+        do_timeout_action();
     }
 
     switch (servant_health) {
         case pcmk_health_pending:
         case pcmk_health_shutdown:
         case pcmk_health_transient:
-            DBGLOG(LOG_INFO, "Not notifying parent: state transient (%d)", servant_health);
+            DBGLOG(LOG_DEBUG, "Not notifying parent: state transient (%d)", servant_health);
             break;
 
         case pcmk_health_unknown:
@@ -901,7 +1126,7 @@ notify_parent(void)
             break;
 
         case pcmk_health_online:
-            DBGLOG(LOG_INFO, "Notifying parent: healthy");
+            DBGLOG(LOG_DEBUG, "Notifying parent: healthy");
             sigqueue(ppid, SIG_LIVENESS, signal_value);
             break;
 
