@@ -16,7 +16,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#include <pacemaker/crm/common/util.h>
+#include <crm/common/util.h>
 #include "sbd.h"
 #define	LOCKSTRLEN	11
 
@@ -31,6 +31,10 @@ int	servant_restart_interval = 5;
 int	servant_restart_count = 1;
 int	start_mode = 0;
 char*	pidfile = NULL;
+bool do_flush = true;
+char timeout_sysrq_char = 'b';
+bool move_to_root_cgroup = true;
+bool enforce_moving_to_root_cgroup = false;
 
 int parse_device_line(const char *line);
 
@@ -40,19 +44,36 @@ void recruit_servant(const char *devname, pid_t pid)
 	struct servants_list_item *newbie;
 
 	if (lookup_servant_by_dev(devname)) {
-		cl_log(LOG_DEBUG, "Servant %s already exists", devname);
-		return;
+	    cl_log(LOG_DEBUG, "Servant %s already exists", devname);
+	    return;
 	}
 
 	newbie = malloc(sizeof(*newbie));
-	if (!newbie) {
-		fprintf(stderr, "malloc failed in recruit_servant.\n");
-		exit(1);
+	if (newbie) {
+	    memset(newbie, 0, sizeof(*newbie));
+	    newbie->devname = strdup(devname);
+	    newbie->pid = pid;
+	    newbie->first_start = 1;
+	}
+	if (!newbie || !newbie->devname) {
+	    fprintf(stderr, "heap allocation failed in recruit_servant.\n");
+	    exit(1);
+	}
+
+	/* some sanity-check on our newbie */
+	if (sbd_is_disk(newbie)) {
+	    cl_log(LOG_INFO, "Monitoring %s", devname);
+	    disk_count++;
+	} else if (sbd_is_pcmk(newbie) || sbd_is_cluster(newbie)) {
+	    /* alive just after pcmk and cluster servants have shown up */
+	    newbie->outdated = 1;
+	} else {
+	    /* toss our newbie */
+	    cl_log(LOG_ERR, "Refusing to recruit unrecognized servant %s", devname);
+	    free((void *) newbie->devname);
+	    free(newbie);
+	    return;
 	}
-	memset(newbie, 0, sizeof(*newbie));
-	newbie->devname = strdup(devname);
-	newbie->pid = pid;
-	newbie->first_start = 1;
 
 	if (!s) {
 		servants_leader = newbie;
@@ -63,12 +84,6 @@ void recruit_servant(const char *devname, pid_t pid)
 	}
 
 	servant_count++;
-        if(sbd_is_disk(newbie)) {
-            cl_log(LOG_NOTICE, "Monitoring %s", devname);
-            disk_count++;
-        } else {
-            newbie->outdated = 1;
-        }
 }
 
 int assign_servant(const char* devname, functionp_t functionp, int mode, const void* argp)
@@ -146,7 +161,7 @@ void servant_start(struct servants_list_item *s)
 	if (sbd_is_disk(s)) {
 #if SUPPORT_SHARED_DISK
 		DBGLOG(LOG_INFO, "Starting servant for device %s", s->devname);
-		s->pid = assign_servant(s->devname, servant, start_mode, s);
+		s->pid = assign_servant(s->devname, servant_md, start_mode, s);
 #else
                 cl_log(LOG_ERR, "Shared disk functionality not supported");
                 return;
@@ -477,19 +492,19 @@ void inquisitor_child(void)
 					if (sbd_is_disk(s)) {
 						if (WIFEXITED(status)) {
 							switch(WEXITSTATUS(status)) {
-								case EXIT_MD_IO_FAIL:
+								case EXIT_MD_SERVANT_IO_FAIL:
 									DBGLOG(LOG_INFO, "Servant for %s requests to be disowned",
 										s->devname);
 									break;
-								case EXIT_MD_REQUEST_RESET:
+								case EXIT_MD_SERVANT_REQUEST_RESET:
 									cl_log(LOG_WARNING, "%s requested a reset", s->devname);
 									do_reset();
 									break;
-								case EXIT_MD_REQUEST_SHUTOFF:
+								case EXIT_MD_SERVANT_REQUEST_SHUTOFF:
 									cl_log(LOG_WARNING, "%s requested a shutoff", s->devname);
 									do_off();
 									break;
-								case EXIT_MD_REQUEST_CRASHDUMP:
+								case EXIT_MD_SERVANT_REQUEST_CRASHDUMP:
 									cl_log(LOG_WARNING, "%s requested a crashdump", s->devname);
 									do_crashdump();
 									break;
@@ -497,6 +512,22 @@ void inquisitor_child(void)
 									break;
 							}
 						}
+					} else if (sbd_is_pcmk(s)) {
+						if (WIFEXITED(status)) {
+							switch(WEXITSTATUS(status)) {
+								case EXIT_PCMK_SERVANT_GRACEFUL_SHUTDOWN:
+									DBGLOG(LOG_INFO, "PCMK-Servant has exited gracefully");
+									/* revert to state prior to pacemaker-detection */
+									s->restarts = 0;
+									s->restart_blocked = 0;
+									cluster_appeared = 0;
+									s->outdated = 1;
+									s->t_last.tv_sec = 0;
+									break;
+								default:
+									break;
+							}
+						}
 					}
 					cleanup_servant_by_pid(pid);
 				}
@@ -565,7 +596,7 @@ void inquisitor_child(void)
                     if(cluster_alive(true)) {
                         /* We LIVE! */
                         if(cluster_appeared == false) {
-                            cl_log(LOG_NOTICE, "Active cluster detected");
+                            cl_log(LOG_INFO, "Active cluster detected");
                         }
                         tickle = 1;
                         can_detach = 1;
@@ -574,7 +605,7 @@ void inquisitor_child(void)
                     } else if(cluster_alive(false)) {
                         if(!decoupled) {
                             /* On the way up, detach and arm the watchdog */
-                            cl_log(LOG_NOTICE, "Partial cluster detected, detaching");
+                            cl_log(LOG_INFO, "Partial cluster detected, detaching");
                         }
 
                         can_detach = 1;
@@ -655,7 +686,7 @@ void inquisitor_child(void)
 				/* At level 2 or above, we do nothing, but expect
 				 * things to eventually return to
 				 * normal. */
-				do_reset();
+				do_timeout_action();
 			} else {
 				cl_log(LOG_ERR, "SBD: DEBUG MODE: Would have fenced due to timeout!");
 			}
@@ -668,7 +699,7 @@ void inquisitor_child(void)
 
                         if (debug_mode && watchdog_use) {
                             /* In debug mode, trigger a reset before the watchdog can panic the machine */
-                            do_reset();
+                            do_timeout_action();
                         }
 		}
 
@@ -751,58 +782,73 @@ int inquisitor(void)
 int
 parse_device_line(const char *line)
 {
-    int lpc = 0;
-    int last = 0;
-    int max = 0;
+    size_t lpc = 0;
+    size_t last = 0;
+    size_t max = 0;
     int found = 0;
+    bool skip_space = true;
+    int space_run = 0;
 
-    if(line) {
-        max = strlen(line);
+    if (!line) {
+        return 0;
     }
 
-    if (max <= 0) {
-        return found;
-    }
+    max = strlen(line);
 
-    cl_log(LOG_DEBUG, "Processing %d bytes: [%s]", max, line);
-    /* Skip initial whitespace */
-    for (lpc = 0; lpc <= max && isspace(line[lpc]); lpc++) {
-        last = lpc + 1;
-    }
+    cl_log(LOG_DEBUG, "Processing %d bytes: [%s]", (int) max, line);
 
-    /* Now the actual content */
     for (lpc = 0; lpc <= max; lpc++) {
-        int a_space = isspace(line[lpc]);
+        if (isspace(line[lpc])) {
+            if (skip_space) {
+                last = lpc + 1;
+            } else {
+                space_run++;
+            }
+            continue;
+        }
+        skip_space = false;
+        if (line[lpc] == ';' || line[lpc] == 0) {
+            int rc = 0;
+            char *entry = calloc(1, 1 + lpc - last);
 
-        if (a_space && lpc < max && isspace(line[lpc + 1])) {
-            /* fast-forward to the end of the spaces */
-
-        } else if (a_space || line[lpc] == ';' || line[lpc] == 0) {
-            int rc = 1;
-            char *entry = NULL;
-
-            if (lpc > last) {
-                entry = calloc(1, 1 + lpc - last);
+            if (entry) {
                 rc = sscanf(line + last, "%[^;]", entry);
+            } else {
+                fprintf(stderr, "Heap allocation failed parsing device-line.\n");
+                exit(1);
             }
 
-            if (entry == NULL) {
-                /* Skip */
-            } else if (rc != 1) {
-                cl_log(LOG_WARNING, "Could not parse (%d %d): %s", last, lpc, line + last);
+            if (rc != 1) {
+                cl_log(LOG_WARNING, "Could not parse: '%s'", line + last);
             } else {
+                entry[strlen(entry)-space_run] = '\0';
                 cl_log(LOG_DEBUG, "Adding '%s'", entry);
                 recruit_servant(entry, 0);
                 found++;
             }
 
             free(entry);
+            skip_space = true;
             last = lpc + 1;
         }
+        space_run = 0;
     }
     return found;
 }
 
+#define SBD_SOURCE_FILES "sbd-cluster.c,sbd-common.c,sbd-inquisitor.c,sbd-md.c,sbd-pacemaker.c,setproctitle.c"
+
+static void
+sbd_log_filter_ctl(const char *files, uint8_t priority)
+{
+	if (files == NULL) {
+		files = SBD_SOURCE_FILES;
+	}
+
+	qb_log_filter_ctl(QB_LOG_SYSLOG, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, files, priority);
+	qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, files, priority);
+}
+
 int
 arg_enabled(int arg_count)
 {
@@ -818,7 +864,9 @@ int main(int argc, char **argv, char **envp)
 	int P_count = 0;
         int qb_facility;
         const char *value = NULL;
-        int start_delay = 0;
+        bool delay_start = false;
+        long delay = 0;
+        char *timeout_action = NULL;
 
 	if ((cmdname = strrchr(argv[0], '/')) == NULL) {
 		cmdname = argv[0];
@@ -834,25 +882,10 @@ int main(int argc, char **argv, char **envp)
 
         qb_log_ctl(QB_LOG_SYSLOG, QB_LOG_CONF_ENABLED, QB_TRUE);
         qb_log_ctl(QB_LOG_STDERR, QB_LOG_CONF_ENABLED, QB_FALSE);
+        sbd_log_filter_ctl(NULL, LOG_NOTICE);
 
 	sbd_get_uname();
 
-        value = getenv("SBD_DEVICE");
-        if(value) {
-#if SUPPORT_SHARED_DISK
-            int devices = parse_device_line(value);
-            if(devices < 1) {
-                fprintf(stderr, "Invalid device line: %s\n", value);
-		exit_status = -2;
-                goto out;
-            }
-#else
-            fprintf(stderr, "Shared disk functionality not supported\n");
-            exit_status = -2;
-            goto out;
-#endif
-        }
-
         value = getenv("SBD_PACEMAKER");
         if(value) {
             check_pcmk = crm_is_true(value);
@@ -899,11 +932,39 @@ int main(int argc, char **argv, char **envp)
 
         value = getenv("SBD_DELAY_START");
         if(value) {
-            start_delay = crm_is_true(value);
-        }
-        cl_log(LOG_DEBUG, "Start delay: %d (%s)", (int)start_delay, value?value:"default");
+            delay_start = crm_is_true(value);
 
-	while ((c = getopt(argc, argv, "czC:DPRTWZhvw:d:n:p:1:2:3:4:5:t:I:F:S:s:")) != -1) {
+            if (!delay_start) {
+                delay = crm_get_msec(value) / 1000;
+                if (delay > 0) {
+                    delay_start = true;
+                }
+            }
+        }
+        cl_log(LOG_DEBUG, "Delay start: %s%s%s",
+               delay_start? "yes (" : "no",
+               delay_start? (delay > 0 ? value: "msgwait") : "",
+               delay_start? ")" : "");
+
+        value = getenv("SBD_TIMEOUT_ACTION");
+        if(value) {
+            timeout_action = strdup(value);
+        }
+
+        value = getenv("SBD_MOVE_TO_ROOT_CGROUP");
+        if(value) {
+            move_to_root_cgroup = crm_is_true(value);
+
+            if (move_to_root_cgroup) {
+               enforce_moving_to_root_cgroup = true;
+            } else {
+                if (strcmp(value, "auto") == 0) {
+                    move_to_root_cgroup = true;
+                }
+            }
+        }
+
+	while ((c = getopt(argc, argv, "czC:DPRTWZhvw:d:n:p:1:2:3:4:5:t:I:F:S:s:r:")) != -1) {
 		switch (c) {
 		case 'D':
 			break;
@@ -926,15 +987,17 @@ int main(int argc, char **argv, char **envp)
 		case 'v':
                     debug++;
                     if(debug == 1) {
-                        qb_log_filter_ctl(QB_LOG_SYSLOG, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, "sbd-common.c,sbd-inquisitor.c,sbd-md.c,sbd-pacemaker.c", LOG_DEBUG);
-                        qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, "sbd-common.c,sbd-inquisitor.c,sbd-md.c,sbd-pacemaker.c", LOG_DEBUG);
-			cl_log(LOG_INFO, "Verbose mode enabled.");
+                        sbd_log_filter_ctl(NULL, LOG_INFO);
+                        cl_log(LOG_INFO, "Verbose mode enabled.");
 
                     } else if(debug == 2) {
+                        sbd_log_filter_ctl(NULL, LOG_DEBUG);
+                        cl_log(LOG_INFO, "Debug mode enabled.");
+
+                    } else if(debug == 3) {
                         /* Go nuts, turn on pacemaker's logging too */
-                        qb_log_filter_ctl(QB_LOG_SYSLOG, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, "*", LOG_DEBUG);
-                        qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD, QB_LOG_FILTER_FILE, "*", LOG_DEBUG);
-			cl_log(LOG_INFO, "Verbose library mode enabled.");
+                        sbd_log_filter_ctl("*", LOG_DEBUG);
+                        cl_log(LOG_INFO, "Debug library mode enabled.");
                     }
                     break;
 		case 'T':
@@ -1016,9 +1079,16 @@ int main(int argc, char **argv, char **envp)
 			cl_log(LOG_INFO, "Servant restart count set to %d",
 					(int)servant_restart_count);
 			break;
+		case 'r':
+			if (timeout_action) {
+				free(timeout_action);
+			}
+			timeout_action = strdup(optarg);
+			break;
 		case 'h':
 			usage();
-			return (0);
+			goto out;
+			break;
 		default:
 			exit_status = -2;
 			goto out;
@@ -1026,6 +1096,28 @@ int main(int argc, char **argv, char **envp)
 		}
 	}
 
+    if (disk_count == 0) {
+        /* if we already have disks from commandline
+           then it is probably undesirable to add those
+           from environment (general rule cmdline has precedence)
+         */
+        value = getenv("SBD_DEVICE");
+        if ((value) && strlen(value)) {
+#if SUPPORT_SHARED_DISK
+            int devices = parse_device_line(value);
+            if(devices < 1) {
+                fprintf(stderr, "Invalid device line: %s\n", value);
+                exit_status = -2;
+                goto out;
+            }
+#else
+            fprintf(stderr, "Shared disk functionality not supported\n");
+            exit_status = -2;
+            goto out;
+#endif
+        }
+	}
+
 	if (watchdogdev == NULL || strcmp(watchdogdev, "/dev/null") == 0) {
             watchdog_use = 0;
 
@@ -1074,6 +1166,37 @@ int main(int argc, char **argv, char **envp)
 		goto out;
 	}
 
+	if (timeout_action) {
+		char *p[2];
+		int i;
+		char c;
+		int nrflags = sscanf(timeout_action, "%m[a-z],%m[a-z]%c", &p[0], &p[1], &c);
+		bool parse_error = (nrflags < 1) || (nrflags > 2);
+
+		for (i = 0; (i < nrflags) && (i < 2); i++) {
+			if (!strcmp(p[i], "reboot")) {
+				timeout_sysrq_char = 'b';
+			} else if (!strcmp(p[i], "crashdump")) {
+				timeout_sysrq_char = 'c';
+			} else if (!strcmp(p[i], "off")) {
+				timeout_sysrq_char = 'o';
+			} else if (!strcmp(p[i], "flush")) {
+				do_flush = true;
+			} else if (!strcmp(p[i], "noflush")) {
+				do_flush = false;
+			} else {
+				parse_error = true;
+			}
+			free(p[i]);
+		}
+		if (parse_error) {
+			fprintf(stderr, "Failed to parse timeout-action \"%s\".\n",
+				timeout_action);
+			exit_status = -1;
+			goto out;
+		}
+	}
+
 #if SUPPORT_SHARED_DISK
 	if (strcmp(argv[optind], "create") == 0) {
 		exit_status = init_devices(servants_leader);
@@ -1099,10 +1222,12 @@ int main(int argc, char **argv, char **envp)
                     open_any_device(servants_leader);
                 }
 
-                if(start_delay) {
-                    unsigned long delay = get_first_msgwait(servants_leader);
+                if (delay_start) {
+                    if (delay <= 0) {
+                        delay = get_first_msgwait(servants_leader);
+                    }
 
-                    sleep(delay);
+                    sleep((unsigned long) delay);
                 }
 
 	} else {
@@ -1132,10 +1257,15 @@ int main(int argc, char **argv, char **envp)
                         recruit_servant("cluster", 0);
                 }
 
+                cl_log(LOG_NOTICE, "%s flush + write \'%c\' to sysrq in case of timeout",
+                       do_flush?"Do":"Skip", timeout_sysrq_char);
                 exit_status = inquisitor();
         }
         
   out:
+	if (timeout_action) {
+				free(timeout_action);
+	}
 	if (exit_status < 0) {
 		if (exit_status == -2) {
 			usage();
